{{
   Retargetable Concurrent Smallc Compiler.
   The RCSC generates P-codes for a virtual machine with two registers and a stack.
   This object interprets these P-codes and updates the registers and stack accordingly.
   Standard input and output are mapped to a serial driver.
   
}}


CON

  vmSTACKSIZE = 256

  {{ comparison attributes }}
  EQ   =    1   {{ 0x01 equal }}
  GE   =  156   {{ 0x9C greater or equal }}
  GEu  =   28   {{ 0x1C greater or equal (unsigned) }}
  GT   =  140   {{ 0x8C greater than }}
  GTu  =   12   {{ 0x0C greater than (unsigned) }}
  LE   =  154   {{ 0x9A less or equal }}
  LEu  =   26   {{ 0x1A less or equal (unsigned) }}
  LT   =  138   {{ 0x8A less than }}
  LTu  =   10   {{ 0x0A less than (unsigned) }}
  NE   =   40   {{ 0x28 not equal }}

  EQf  =    1   {{ 0x01 equal }}
  GEf  =  156   {{ 0x9C greater or equal }}
  GTf  =  140   {{ 0x8C greater than }}
  LEf  =  154   {{ 0x9A less or equal }}
  LTf  =  138   {{ 0x8A less than }}
  NEf  =   40   {{ 0x28 not equal }}

  {{ pcode followed by 0 operands }}
  PCODE_OPER0 = 0
  pNOP     =   0   {{ no operation }}
  pADD12   =   1   {{ add sr to pr }}
  pADD21   =   2   {{ add pr to sr }}
  pAND12   =   3   {{ AND sr to pr }}
  pANEG1   =   4   {{ arith negate pr }}
  pASL12   =   5   {{ arith shift left sr by pr into pr }}
  pASR12   =   6   {{ arith shift right sr by pr into pr }}
  pCALL1   =   7   {{ call function thru pr }}
  pCLI     =   8   {{ interrupt disable }}
  pCOM1    =   9   {{ ones complement pr }}
  pCSCINIT =  10   {{ concurrent Small C Initialization }}
  pDBL1    =  11   {{ double pr }}
  pDBL2    =  12   {{ double sr }}
  pDIV12   =  13   {{ div pr by sr }}
  pDIV12u  =  14   {{ div pr by sr unsigned }}
  pENTER   =  15   {{ set stack frame on function entry }}
  pLNEG1   =  16   {{ logical negate pr }}
  pMOD12   =  17   {{ modulo pr by sr }}
  pMOD12u  =  18   {{ modulo pr by sr unsigned }}
  pMOVE21  =  19   {{ move pr to sr }}
  pMUL12   =  20   {{ multiply pr by sr }}
  pMUL12u  =  21   {{ multiply pr by sr unsigned }}
  pOR12    =  22   {{ OR sr onto pr }}
  pPOP2    =  23   {{ pop stack into sr }}
  pPUSH1   =  24   {{ push pr onto stack }}
  pPUSH2   =  25   {{ push sr }}
  pPUTbp1  =  26   {{ put pr byte in mem thru sr ptr }}
  pPUTwp1  =  27   {{ put pr word in mem thru sr ptr }}
  pRETURN  =  28   {{ restore stack and return }}
  pSTI     =  29   {{ interrupt enable }}
  pSUB12   =  30   {{ sub sr from pr }}
  pSWAP12  =  31   {{ swap pr and sr }}
  pSWAP1s  =  32   {{ swap pr and top of stack }}
  pSWITCH  =  33   {{ find switch case }}
  pXOR12   =  34   {{ XOR pr with sr }}

  {{ pcode followed by 1 operand }}
  PCODE_OPER1 = 35
  pADDbpn  =  35   {{ add n to mem byte thru sr ptr }}
  pARGCNTn =  36   {{ pass arg count to function }}
  pCMP12   =  37   {{ set pr to (sr CMP pr) }}

  {{ pcode followed by 2 operands }}
  PCODE_OPER2 = 38
  pADD1n   =  38   {{ add n to pr }}
  pADD2n   =  39   {{ add immediate to sr }}
  pADDSP   =  40   {{ add to stack pointer }}
  pADDwpn  =  41   {{ add n to mem word thru sr ptr }}
  pANDbm1  =  42   {{ bitwise and pr with mem byte thru label }}
  pANDwm1  =  43   {{ bitwise and pr with mem word thru label }}
  pCALLm   =  44   {{ call function directly }}
  pGETb1m  =  45   {{ get byte into pr from mem thru label }}
  pGETb1mu =  46   {{ get unsigned byte into pr from mem thru label }}
  pGETb1p  =  47   {{ get byte into pr from mem thru sr ptr }}
  pGETb1pu =  48   {{ get unsigned byte into pr from mem thru sr ptr }}
  pGETb1s  =  49   {{ get byte into pr from stack }}
  pGETb1su =  50   {{ get unsigned byte into pr from stack }}
  pGETw1m  =  51   {{ get word into pr from mem thru label }}
  pGETw1n  =  52   {{ get word of value n into pr }}
  pGETw1p  =  53   {{ get word into pr from mem thru sr ptr }}
  pGETw1s  =  54   {{ get word into pr from stack }}
  pGETw2m  =  55   {{ get word into sr from mem (label) }}
  pGETw2n  =  56   {{ get word of value n into sr }}
  pGETw2p  =  57   {{ get word into sr thru sr ptr }}
  pGETw2s  =  58   {{ get word into sr from stack }}
  pJMPm    =  59   {{ jump to label }}
  pORbm1   =  60   {{ bitwise or pr with mem byte thru label }}
  pORwm1   =  61   {{ bitwise or pr with mem word thru label }}
  pPOINT1m =  62   {{ point pr to mem item thru label }}
  pPOINT1s =  63   {{ point pr to stack item }}
  pPOINT2m =  64   {{ point sr to mem thru label }}
  pPOINT2s =  65   {{ point sr to stack }}
  pPUSHm   =  66   {{ push word from mem thru label }}
  pPUSHp   =  67   {{ push word from mem thru sr ptr }}
  pPUSHs   =  68   {{ push word from stack }}
  pPUTbm1  =  69   {{ put pr byte in mem thru label }}
  pPUTwm1  =  70   {{ put pr word in mem thru label }}
  pXORbm1  =  71   {{ bitwise xor pr with mem byte thru label }}
  pXORwm1  =  72   {{ bitwise xor pr with mem word thru label }}

  {{ pcode followed by 3 operands }}
  PCODE_OPER3 = 73
  pADDbmn  =  73   {{ add n to mem byte thru label }}
  pANDbmn  =  74   {{ bitwise and n with mem byte thru label }}
  pCMP10f  =  75   {{ jump to label if (pr CMP 0) is false }}
  pORbmn   =  76   {{ bitwise or n with mem byte thru label }}
  pPUTbmn  =  77   {{ put n into mem byte thru label }}
  pXORbmn  =  78   {{ bitwise xor n with mem byte thru label }}

  {{ pcode followed by 4 operands }}
  PCODE_OPER4 = 79
  pADDwmn  =  79   {{ add n to mem word thru label }}
  pANDwmn  =  80   {{ bitwise and n with mem word thru label }}
  pORwmn   =  81   {{ bitwise or n with mem word thru label }}
  pPUTwmn  =  82   {{ put n into mem word thru label }}
  pXORwmn  =  83   {{ bitwise xor n with mem word thru label }}


DAT

  {{ math registers }}
  vmAXlong      long    0               {{ for multiply, divide and modulo }}
  vmBXlong      long    0               {{ for multiply, divide and modulo }}

  {{ the virtual machine registers }}
  vmAX          word    0               {{ primary register (pr)           }}
  vmBX          word    0               {{ secondary register (sr)         }}
  vmSP          word    0               {{ stack pointer register          }}

  {{ working registers }}
  vmCX          word    0               {{ auxiliary register #1           }}
  vmDX          word    0               {{ auxiliary register #2           }}
  vmEX          word    0               {{ auxiliary register #3           }}
  vmPC          word    0               {{ program counter                 }}
  vmBP          word    0               {{ stack frame register            }}
  vmSTACKbottom word                    {{ lowest value of vmSP            }}
  vmSTACKtop    word                    {{ highest value of vmSP           }}
  vmPCODE       byte    0               {{ pseude code operator            }}
  vmARGCNT      byte    0               {{ argument count                  }}
  vmINTERRUPT   byte    0               {{ tue if interrupt enabled        }}
  vmSTACK       byte    0[vmSTACKSIZE]  {{ stack area                      }}
  

PUB reset(addr)
  '/**
  ' * Reset program counter and initialize stack.
  ' *
  ' * @param addr Start address (main of C program).
  ' */
  vmSP := @vmSTACK + vmSTACKSIZE  {{initialize stack }}
  vmSTACKtop := vmSP
  vmSTACKbottom := @vmSTACK
  vmPC := addr                    {{initialize program counter }}


PUB run(pcodes)
  '/**
  ' * Run a number of P-codes
  ' *
  ' * @param pcodes Number of P-codes to execute. Use 0 to run forever.
  ' */
  if pcodes > 0
    repeat pcodes
      nextPcode
  else
    repeat
      nextPcode
      

PUB getPR: pr
  '/**
  ' * Get primary register.
  ' *
  ' * @return Primary register.
  ' */
  return vmAX


PUB getSR: sr
  '/**
  ' * Get secondary register.
  ' *
  ' * @return Secondary register.
  ' */
  return vmBX


PUB getSP: sp
  '/**
  ' * Get stack pointer.
  ' *
  ' * @return Stack pointer.
  ' */
  return vmSP


PUB stackOK: YesNo
  '/**
  ' * Check stack and memorize lowest and highest stack pointer values.
  ' *
  ' * @return True if stack pointer in valid area, false if stack pointer ever moved outside stack area.
  ' */
  if vmSP < vmSTACKbottom
    vmSTACKbottom := vmSP
  if vmSP > vmSTACKtop
    vmSTACKtop := vmSP
  return (vmSTACKbottom => @vmSTACK) AND (vmSTACKtop =< @vmSTACK + vmSTACKSIZE)


PUB getLowestSP: bottom
  '/**
  ' * Get lowest stack pointer value.
  ' *
  ' * @return Lowest stack pointer value.
  ' */
  return vmSTACKbottom
     
  
PUB getHighestSP: top
  '/**
  ' * Get highest stack pointer value.
  ' *
  ' * @return Highest stack pointer value.
  ' */
  return vmSTACKtop
     
  
PUB nextPcode | sign
  '/**
  ' * Execute next P-code.
  ' */
  vmPCODE := byte[vmPC++]              {{ pcode }}
  if vmPCODE => PCODE_OPER1
    vmCX.byte[0] := byte[vmPC++]       {{ operand 1 }}
    if vmPCODE => PCODE_OPER2
      vmCX.byte[1] := byte[vmPC++]     {{ operand 2 }}
      if vmPCODE => PCODE_OPER3
        vmDX.byte[0] := byte[vmPC++]   {{ operand 3 }}
        if vmPCODE => PCODE_OPER4
          vmDX.byte[1] := byte[vmPC++] {{ operand 4 }}
          {{ pcode operators with 4 operands }}
          case vmPCODE
            pADDwmn:  {{ add n to memory word thru label }}
                      {{ addr in CX, n in DX }}
                      vmEX.byte[0] := byte[vmCX]
                      vmEX.byte[1] := byte[vmCX+1]
                      vmDX += vmEX
                      byte[vmCX] := vmDX.byte[0]
                      byte[vmCX+1] := vmDX.byte[1]

            pANDwmn:  {{ bitwise and n with mem word thru label }}
                      {{ addr in CX, n in DX }}
                      byte[vmCX] &= vmDX.byte[0]
                      byte[vmCX+1] &= vmDX.byte[1]

            pORwmn:   {{ bitwise or n with mem word thru label }}
                      {{ addr in CX, n in DX }}
                      byte[vmCX] |= vmDX.byte[0]
                      byte[vmCX+1] |= vmDX.byte[1]

            pPUTwmn:  {{ put n into memory word thru label }}
                      {{ addr in CX, n in DX }}
                      byte[vmCX] := vmDX.byte[0]
                      byte[vmCX+1] := vmDX.byte[1]

            pXORwmn:  {{ bitwise xor n with mem word thru label }}
                      {{ addr in CX, n in DX }}
                      byte[vmCX] ^= vmDX.byte[0]
                      byte[vmCX+1] ^= vmDX.byte[1]

        else
          {{ pcode operators with 3 operands }}
          case vmPCODE
            pADDbmn:  {{ add n to memory byte thru label }}
                      {{ addr in CX, n in DL }}
                      byte[vmCX] += vmDX.byte[0]

            pANDbmn:  {{ bitwise and n with memory byte }}
                      {{ addr in CX, n in DL }}
                      byte[vmCX] &= vmDX.byte[0]

            pCMP10f:  {{ jump to label if (pr CMP 0) is false }}
                      {{ addr in CX, cmp in DL }}
                      case vmDX.byte[0]
                        EQf: {{ equal }}
                             vmEX := (vmAX == 0)
                        GEf: {{ greater or equal }}
                             vmEX := (vmAX => 0)
                        GTf: {{ greater }}
                             vmEX := (vmAX > 0)
                        LEf: {{ less or equal }}
                             vmEx := (vmAX =< 0)
                        LTf: {{ less }}
                             vmEX := (vmAX < 0)
                        NEf: {{ not equal }}
                             vmEX := (vmAX <> 0)
                        other: vmEX := true
                      if NOT vmEX
                        vmPC := vmCX

            pORbmn:   {{ bitwise or n with memory byte }}
                      {{ addr in CX, n in DL }}
                      byte[vmCX] |= vmDX.byte[0]

            pPUTbmn:  {{ put n into memory byte thru label }}
                      {{ addr in CX, n in DL }}
                      byte[vmCX] := vmDX.byte[0]

            pXORbmn:  {{ bitwise xor n with memory byte }}
                      {{ addr in CX, n in DL }}
                      byte[vmCX] ^= vmDX.byte[0]

      else
        {{ pcode operators with 2 operands }}
        case vmPCODE
          pADD1n:   {{ add n to pr }}
                    {{ n in CX }}
                    vmAX += vmCX

          pADD2n:   {{ add n to sr }}
                    {{ n in CX }}
                    vmBX += vmCX

          pADDSP:   {{ add n to stackpointer }}
                    {{ n in CX }}
                    vmSP += vmCX

          pADDwpn:  {{ add n to memory word thru sr pointer }}
                    {{ n in CX }}
                    vmEX.byte[0] := byte[vmBX]
                    vmEX.byte[1] := byte[vmBX+1]
                    vmCX += vmEX
                    byte[vmBX] := vmCX.byte[0]
                    byte[vmBX+1] := vmCX.byte[1]

          pANDbm1:  {{ bitwise and pr with mem byte thru label }}
                    {{ addr in CX }}
                    byte[vmCX] &= vmAX.byte[0]

          pANDwm1:  {{ bitwise and pr with mem word thru label }}
                    {{ addr in CX }}
                    byte[vmCX] &= vmAX.byte[0]
                    byte[vmCX+1] &= vmAX.byte[1]

          pCALLm:   {{ call function directly }}
                    {{ addr in CX }}
                    byte[--vmSP] := vmPC.byte[1]
                    byte[--vmSP] := vmPC.byte[0]
                    vmPC := vmCX

          pGETb1m:  {{ get byte into pr from memory thru label }}
                    {{ addr in CX }}
                    vmAX.byte[0] := byte[vmCX]
                    vmAX.byte[1] := (vmAX.byte[0] & $80) <> 0

          pGETb1mu: {{ get unsigned byte into pr from memory thru label }}
                    {{ addr in CX }}
                    vmAX.byte[0] := byte[vmCX]
                    vmAX.byte[1] := 0

          pGETb1p:  {{ get byte into pr from memory thru sr pointer }}
                    {{ offset in CX }}
                    vmAX.byte[0] := byte[vmBX+vmCX]
                    vmAX.byte[1] := (vmAX.byte[0] & $80) <> 0
                  
          pGETb1pu: {{ get unsigned byte into pr from memory thru sr pointer }}
                    {{ offset in CX }}
                    vmAX.byte[0] := byte[vmBX+vmCX]
                    vmAX.byte[1] := 0

          pGETb1s:  {{ get byte into pr from stack }}
                    {{ offset in CX }}
                    vmAX.byte[0] := byte[vmBP+vmCX]
                    vmAX.byte[1] := (vmAX.byte[0] & $80) <> 0
                  
          pGETb1su: {{ get unsigned byte into pr from stack }}
                    {{ offset in CX }}
                    vmAX.byte[0] := byte[vmBP+vmCX]
                    vmAX.byte[1] := 0

          pGETw1m:  {{ get word into pr from memory thru label }}
                    {{ addr in CX }}
                    vmAX.byte[0] := byte[vmCX]
                    vmAX.byte[1] := byte[vmCX+1]

          pGETw1n:  {{ get word of value n into pr }}
                    {{ n in CX }}
                    vmAX := vmCX

          pGETw1p:  {{ get word into pr from memory thru sr pointer }}
                    {{ offset in CX }}
                    vmAX.byte[0] := byte[vmBX+vmCX]
                    vmAX.byte[1] := byte[vmBX+vmCX+1]
                  
          pGETw1s:  {{ get word into pr from stack }}
                    {{ offset in CX }}
                    vmAX.byte[0] := byte[vmBP+vmCX]
                    vmAX.byte[1] := byte[vmBP+vmCX+1]
                  
          pGETw2m:  {{ get word into sr from memory thru label }}
                    {{ addr in CX }}
                    vmBX.byte[0] := byte[vmCX]
                    vmBX.byte[1] := byte[vmCX+1]

          pGETw2n:  {{ get word of value n into sr }}
                    {{ n in CX }}
                    vmBX := vmCX

          pGETw2p:  {{ get word into sr from memory thru sr pointer }}
                    {{ offset in CX }}
                    vmDX.byte[0] := byte[vmBX+vmCX]
                    vmDX.byte[1] := byte[vmBX+vmCX+1]
                    vmBX := vmDX
                  
          pGETw2s:  {{ get word into sr from stack }}
                    {{ offset in CX }}
                    vmBX.byte[0] := byte[vmBP+vmCX]
                    vmBX.byte[1] := byte[vmBP+vmCX+1]
                  
          pJMPm:    {{ jump to label }}
                    {{ addr in CX }}
                    vmPC := vmCX

          pORbm1:   {{ bitwise or pr with mem byte thru label }}
                    {{ addr in CX }}
                    byte[vmCX] |= vmAX.byte[0]
                    
          pORwm1:   {{ bitwise or pr with mem word thru label }}
                    {{ addr in CX }}
                    byte[vmCX] |= vmAX.byte[0]
                    byte[vmCX+1] |= vmAX.byte[1]

          pPOINT1m: {{ point pr to memory item thru label }}
                    {{ addr in CX }}
                    vmAX := vmCX
                    
          pPOINT1s: {{ point pr to stack item }}
                    {{ offset in CX }}
                    vmAX := vmBP+vmCX
                    
          pPOINT2m: {{ point sr to memory item thru label }}
                    {{ addr in CX }}
                    vmBX := vmCX
                    
          pPOINT2s: {{ point sr to stack item }}
                    {{ offset in CX }}
                    vmBX := vmBP+vmCX
                    
          pPUSHm:   {{ push word from memory thru label }}
                    {{ addr in CX }}
                    byte[--vmSP] := byte[vmCX]
                    byte[--vmSP] := byte[vmCX+1]
                    
          pPUSHp:   {{ push word from memory thru sr pointer }}
                    {{ offset in CX }}
                    byte[--vmSP] := byte[vmBX+vmCX]
                    byte[--vmSP] := byte[vmBX+vmCX+1]
                    
          pPUSHs:   {{ push word from stack }}
                    {{ offset in CX }}
                    byte[--vmSP] := byte[vmBP+vmCX]
                    byte[--vmSP] := byte[vmBP+vmCX+1]
                    
          pPUTbm1:  {{ put pr byte into memory thru label }}
                    {{ addr in CX }}
                    byte[vmCX] := vmAX.byte[0]
                    
          pPUTwm1:  {{ put pr word into memory thru label }}
                    {{ addr in CX }}
                    byte[vmCX] := vmAX.byte[0]
                    byte[vmCX+1] := vmAX.byte[1]
                    
          pXORbm1:  {{ bitwise xor pr with mem byte thru label }}
                    {{ addr in CX }}
                    byte[vmCX] ^= vmAX.byte[0]

          pXORwm1:  {{ bitwise xor pr with mem word thru label }}
                    {{ addr in CX }}
                    byte[vmCX] ^= vmAX.byte[0]
                    byte[vmCX+1] ^= vmAX.byte[1]

    else
      {{ pcode operators with 1 operand }}
      case vmPCODE
        pADDbpn:  {{ add n to memory byte through sr pointer }}
                  {{ n in CL }}
                  byte[vmBX] += vmCX.byte[0]
                  
        pARGCNTn: {{ pass argument count to function }}
                  {{ argcnt in CL }}
                  vmARGCNT := vmCX.byte[0]
                  
        pCMP12:   {{ set pr to (sr CMP pr) }}
                  {{ cmp in CL }}
                  case vmCX.byte[0]
                    EQ : {{ 0x01 equal }}
                         vmAX := (vmBX == vmAX)
                    GE : {{ 0x9C greater or equal }}
                         if (vmBX == vmAX)
                           vmAX := true
                         else
                           vmAX := ((vmBX - vmAX) & $8000) == 0 
                    GEu: {{ 0x1C greater or equal (unsigned) }}
                         vmAX := (vmBX => vmAX)
                    GT : {{ 0x8C greater than }}
                         vmAX := ((vmBX - vmAX) & $8000) == 0
                    GTu: {{ 0x0C greater than (unsigned) }}
                         vmAX := (vmAX > vmBX)
                    LE : {{ 0x9A less or equal }}
                         if (vmBX == vmAX)
                           vmAX := true
                         else
                           vmAX := ((vmBX - vmAX) & $8000) <> 0
                    LEu: {{ 0x1A less or equal (unsigned) }}
                         vmAX := (vmBX =< vmAX)
                    LT : {{ 0x8A less than }}
                         vmAX := ((vmBX - vmAX) & $8000) <> 0
                    LTu: {{ 0x0A less than (unsigned) }}
                         vmAX := (vmBX < vmAX)
                    NE : {{ 0x28 not equal }}
                         vmAX := (vmAX <> vmBX)
                       
  else
    {{ pcode operators with 0 operands }}
    case vmPCODE
      pNOP:     {{ no operation }}
                {{ do nothing }}
                
      pADD12:   {{ add sr to pr }}
                vmAX += vmBX
                
      pADD21:   {{ add pr to sr }}
                vmBX += vmAX
                
      pAND12:   {{ and pr with sr }}
                vmAX &= vmBX
                
      pANEG1:   {{ arithmetically negate pr }}
                vmAX := -vmAX
                
      pASL12:   {{ arithmetically shift left sr by pr into pr }}
                vmAX := (vmBX << vmAX)
                
      pASR12:   {{ arithmetically shift right sr by pr into pr }}
                vmAX := (vmBX -> vmAX)
                
      pCALL1:   {{ call function thru pr }}
                byte[--vmSP] := vmPC.byte[1]
                byte[--vmSP] := vmPC.byte[0]
                vmPC := vmAX
                
      pCLI:     {{ disable interrupt }}
                vmINTERRUPT := false
      
      pCOM1:    {{ one complement pr }}
                vmAX ^= $FFFF
                
      pCSCINIT: {{ initialize csc }}
      
      pDBL1:    {{ double pr }}
                vmAX <<= 1
                
      pDBL2:    {{ double sr }}
                vmBX <<= 1
                
      pDIV12:   {{ divide pr by sr }}
                vmAXlong := ||(~~vmAX) {{sign extend b15, then absolute value }}
                vmBXlong := ||(~~vmBX) {{sign extend b15, then absolute value }}
                sign := ((vmAX ^ vmBX) & $8000) == 0
                vmAXlong /= vmBXlong
                if sign
                  vmAXlong := -vmAXlong
                vmAX := vmAXlong.byte[0]

      pDIV12u:  {{ divide pr by sr (unsigned) }}
                vmAXlong := vmAX
                vmBXlong := vmBX
                vmAXlong /= vmBXlong
                vmAX := vmAXlong.byte[0]

      pENTER:   {{ set stack frame upon function entry }}
                byte[--vmSP] := vmBP.byte[1]
                byte[--vmSP] := vmBP.byte[0]
                vmBP := vmSP
                
      pLNEG1:   {{ logical negation of pr }}
                vmAX := NOT vmAX
                
      pMOD12:   {{ modulo pr by sr }}
                vmAXlong := ||(~~vmAX) {{sign extend b15, then absolute value }}
                vmBXlong := ||(~~vmBX) {{sign extend b15, then absolute value }}
                sign := ((vmAX ^ vmBX) & $8000) == 0
                vmAXlong //= vmBXlong
                if sign
                  vmAXlong := -vmAXlong
                vmAX := vmAXlong.byte[0]
                
      pMOD12u:  {{ modulo pr by sr (unsigned) }}
                vmAXlong := vmAX
                vmBXlong := vmBX
                vmAXlong //= vmBXlong
                vmAX := vmAXlong.byte[0]
      
      pMOVE21:  {{ move pr to sr }}
                vmBX := vmAX
                
      pMUL12:   {{ multiply pr by sr }}
                vmAXlong := ||(~~vmAX) {{sign extend b15, then absolute value }}
                vmBXlong := ||(~~vmBX) {{sign extend b15, then absolute value }}
                sign := ((vmAX ^ vmBX) & $8000) == 0
                vmAXlong *= vmBXlong
                if sign
                  vmAXlong := -vmAXlong
                vmAX := vmAXlong.byte[0]

      pMUL12u:  {{ multiply pr by sr (unsigned) }}
                vmAXlong := vmAX
                vmBXlong := vmBX
                vmAXlong *= vmBXlong
                vmAX := vmAXlong.byte[0]
      
      pOR12:    {{ or pr with sr }}
                vmAX |= vmBX
                
      pPOP2:    {{ pop stack into sr }}
                vmBX.byte[0] := byte[vmSP++]
                vmBX.byte[1] := byte[vmSP++]
                
      pPUSH1:   {{ push pr onto stack }}
                byte[--vmSP] := vmAX.byte[1]
                byte[--vmSP] := vmAX.byte[0]
                
      pPUSH2:   {{ push sr onto stack }}
                byte[--vmSP] := vmBX.byte[1]
                byte[--vmSP] := vmBX.byte[0]
                
      pPUTbp1:  {{ put pr byte into memory thru sr pointer }}
                byte[vmBX] := vmAX.byte[0]
                
      pPUTwp1:  {{ put pr word into memory thru sr pointer }}
                byte[vmBX] := vmAX.byte[0]
                byte[vmBX+1] := vmAX.byte[1]
                
      pRETURN:  {{ restore stack and return }}
                vmSP := vmBP
                vmBP.byte[0] := byte[vmSP++]
                vmBP.byte[1] := byte[vmSP++]
                
      pSTI:     {{ enable interrupt }}
                vmINTERRUPT := true
      
      pSUB12:   {{ subtract sr from pr }}
                vmAX -= vmBX
                
      pSWAP12:  {{ swap pr and sr }}
                vmCX := vmAX
                vmAX := vmBX
                vmBX := vmCX
                
      pSWAP1s:  {{ swap pr and top of stack }}
                vmCX.byte[0] := byte[vmSP++]
                vmCX.byte[1] := byte[vmSP++]
                byte[--vmSP] := vmAX.byte[1]
                byte[--vmSP] := vmAX.byte[0]
                vmAX := vmCX
                
      pSWITCH:  {{ scan switch table to find case }}
                {{ remember: AX=compare value, PC=table address upon entry }}
                {{ the table consists of address/value pairs, a zero address }}
                {{ terminates the table and is followed by a jump to the default prefix }}
                repeat
                  vmCX.byte[0] := byte[vmPC++]   {{ load jump address for case }}
                  vmCX.byte[1] := byte[vmPC++]
                  if vmCX <> 0                   {{ if addr == 0 then do default }}
                    vmDX.byte[0] := byte[vmPC++] {{ get case value }}
                    vmDX.byte[1] := byte[vmPC++]
                    if vmAX == vmDX              {{ compare with compare value }}
                      vmPC := vmCX
                      vmCX := 0
                until vmCX == 0
      
      pXOR12:   {{ xor pr with sr }}
                vmAX ^= vmBX

